# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_hub/module_attachment.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow_hub/module_attachment.proto',
  package='tensorflow_hub',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n&tensorflow_hub/module_attachment.proto\x12\x0etensorflow_hub\".\n\x10ModuleAttachment\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x62\x06proto3')
)




_MODULEATTACHMENT = _descriptor.Descriptor(
  name='ModuleAttachment',
  full_name='tensorflow_hub.ModuleAttachment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='tensorflow_hub.ModuleAttachment.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='tensorflow_hub.ModuleAttachment.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=58,
  serialized_end=104,
)

DESCRIPTOR.message_types_by_name['ModuleAttachment'] = _MODULEATTACHMENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ModuleAttachment = _reflection.GeneratedProtocolMessageType('ModuleAttachment', (_message.Message,), dict(
  DESCRIPTOR = _MODULEATTACHMENT,
  __module__ = 'tensorflow_hub.module_attachment_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow_hub.ModuleAttachment)
  ))
_sym_db.RegisterMessage(ModuleAttachment)


# @@protoc_insertion_point(module_scope)
