# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow_hub/image_module_info.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow_hub/image_module_info.proto',
  package='tensorflow_hub',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n&tensorflow_hub/image_module_info.proto\x12\x0etensorflow_hub\"\x8a\x01\n\x0fImageModuleInfo\x12\x45\n\x12\x64\x65\x66\x61ult_image_size\x18\x01 \x01(\x0b\x32).tensorflow_hub.ImageModuleInfo.InputSize\x1a\x30\n\tInputSize\x12\x0e\n\x06height\x18\x01 \x01(\x05\x12\r\n\x05width\x18\x02 \x01(\x05J\x04\x08\x03\x10\x04\x62\x06proto3')
)




_IMAGEMODULEINFO_INPUTSIZE = _descriptor.Descriptor(
  name='InputSize',
  full_name='tensorflow_hub.ImageModuleInfo.InputSize',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='height', full_name='tensorflow_hub.ImageModuleInfo.InputSize.height', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='tensorflow_hub.ImageModuleInfo.InputSize.width', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=149,
  serialized_end=197,
)

_IMAGEMODULEINFO = _descriptor.Descriptor(
  name='ImageModuleInfo',
  full_name='tensorflow_hub.ImageModuleInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='default_image_size', full_name='tensorflow_hub.ImageModuleInfo.default_image_size', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_IMAGEMODULEINFO_INPUTSIZE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=59,
  serialized_end=197,
)

_IMAGEMODULEINFO_INPUTSIZE.containing_type = _IMAGEMODULEINFO
_IMAGEMODULEINFO.fields_by_name['default_image_size'].message_type = _IMAGEMODULEINFO_INPUTSIZE
DESCRIPTOR.message_types_by_name['ImageModuleInfo'] = _IMAGEMODULEINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ImageModuleInfo = _reflection.GeneratedProtocolMessageType('ImageModuleInfo', (_message.Message,), dict(

  InputSize = _reflection.GeneratedProtocolMessageType('InputSize', (_message.Message,), dict(
    DESCRIPTOR = _IMAGEMODULEINFO_INPUTSIZE,
    __module__ = 'tensorflow_hub.image_module_info_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow_hub.ImageModuleInfo.InputSize)
    ))
  ,
  DESCRIPTOR = _IMAGEMODULEINFO,
  __module__ = 'tensorflow_hub.image_module_info_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow_hub.ImageModuleInfo)
  ))
_sym_db.RegisterMessage(ImageModuleInfo)
_sym_db.RegisterMessage(ImageModuleInfo.InputSize)


# @@protoc_insertion_point(module_scope)
